#include <iostream>
#include <string>
#include <cstring>
#include <climits>
#include <vector>
#include <algorithm>
#include <cassert>
#include <fstream>
#include <stdexcept>
#include <sstream>
#include "skgrid.hpp"


/* The ``double-edge'' logic that I lastly added gains actually nothing.
 * Another borderline logic which could help, in the same spirit of
 * double-edge, is when within two parallel sets (parallel rows/cols, or
 * two squares crossed by common rows/cols), we have edges which actually
 * collide. This implies that the corresponfing number may be killed in
 * the orthogonal direction.
 *
 * Well, I doubt that it would gain much.
 */
using namespace std;

bool no_guess=false;

/* {{{ utility conversion functions */
unsigned int cell_fromrow(unsigned int row, unsigned int idx)
{
    return row*9+idx;
}

unsigned int cell_fromcol(unsigned int col, unsigned int idx)
{
    return idx*9+col;
}

unsigned int cell_fromsqr(unsigned int sqr, unsigned int idx)
{
    unsigned int s1 = sqr%3;
    unsigned int i0 = idx/3;
    unsigned int row = (sqr - s1)+i0;
    unsigned int col = s1*3 + (idx % 3);
    return row*9+col;
}

void row_fromcell(unsigned int& row, unsigned int &idx, unsigned int p)
{
    row = p / 9;
    idx = p % 9;
}


void col_fromcell(unsigned int& col, unsigned int &idx, unsigned int p)
{
    idx = p / 9;
    col = p % 9;
}

void rcs_fromcell(
        std::pair<unsigned int, unsigned int> & r,
        std::pair<unsigned int, unsigned int> & c,
        std::pair<unsigned int, unsigned int> & s,
        unsigned int p)
{
    unsigned int row = p / 9;
    unsigned int col = p % 9;
    unsigned int r1 = row%3;
    unsigned int c1 = col%3;
    unsigned int sqr = row-r1 + col/3;
    unsigned int idx = r1*3+c1;

    r.first = row; r.second = col;
    c.first = col; c.second = row;
    s.first = sqr; s.second = idx;
}
void sqr_fromcell(unsigned int& sqr, unsigned int &idx, unsigned int p)
{
    std::pair<unsigned int, unsigned int> r,c,s;
    rcs_fromcell(r,c,s,p);
    sqr = s.first;
    idx = s.second;
}
/* }}} */
static inline bool is_bit(uint16_t x) {
    return (x & (x-1)) == 0;
}

unsigned int solcount=0;

string allbits(uint16_t t, string sep = " ")
{
    std::ostringstream o;
    for(unsigned int s = 0 ; t ; s++) {
        unsigned int a = __builtin_ffs(t)-1;
        t ^= 1 << a;
        if (s) o << sep;
        o << (a+1);
    }
    return o.str();
}

struct game {

uint16_t cell_whofits[81];

/* possible placements for digits in rows, columns, squares */
uint16_t ppl_row[9][9];
uint16_t ppl_col[9][9];
uint16_t ppl_sqr[9][9];

/* order in which the grid has been filled */
unsigned int pos[81];
unsigned int val[81];
bool is_set[81];
bool rset[9][9];
bool cset[9][9];
bool sset[9][9];

bool restrict_rs[9][9];
bool restrict_cs[9][9];
bool restrict_sr[9][9];
bool restrict_sc[9][9];

unsigned int npos;



void initialize()
{
    npos=0;
    for (unsigned int i = 0 ; i < 9 ; i++) {
        for (unsigned int j = 0 ; j < 9 ; j++) {
            ppl_row[i][j]=0x1ff;
            ppl_col[i][j]=0x1ff;
            ppl_sqr[i][j]=0x1ff;
            rset[i][j]=false;
            cset[i][j]=false;
            sset[i][j]=false;
            restrict_rs[i][j]=false;
            restrict_cs[i][j]=false;
            restrict_sr[i][j]=false;
            restrict_sc[i][j]=false;
        }
    }
    for (unsigned int i = 0 ; i < 81 ; i++) {
        cell_whofits[i]=0x1ff;
        is_set[i]=false;
    }

}

void forbid_pos(unsigned int p, unsigned int x)
{
    cell_whofits[p] &= ~(1<<x);
}

/* Sets a value for cell at position p. */
void set_pos(unsigned int p, unsigned int x)
{
    assert(!is_set[p]);
    if ((cell_whofits[p] >> x & 1) == 0) {
        std::ostringstream o;
        o << "position " << (1+p) << " cannot be filled";
        throw std::logic_error(o.str());
    }

    is_set[p] = true;
    pos[npos]=p;
    val[npos]=x;
    npos++;

    std::pair<unsigned int, unsigned int> r,c,s;
    rcs_fromcell(r,c,s,p);

    for(unsigned int i = 0 ; i < 9 ; i++) {
        ppl_row[r.first][i] &= ~(1 << r.second);
        ppl_col[c.first][i] &= ~(1 << c.second);
        ppl_sqr[s.first][i] &= ~(1 << s.second);
    }

    for(unsigned int i = 0 ; i < 9 ; i++) {
        /* kill in row */
        cell_whofits[cell_fromrow(r.first, i)] &= ~(1 << x);
        /* kill in col */
        cell_whofits[cell_fromcol(c.first, i)] &= ~(1 << x);
        /* kill in sqr */
        cell_whofits[cell_fromsqr(s.first, i)] &= ~(1 << x);
    }

    cell_whofits[p] = 1 << x;
    ppl_row[r.first][x] = 1 << r.second;
    ppl_col[c.first][x] = 1 << c.second;
    ppl_sqr[s.first][x] = 1 << s.second;

    rset[r.first][x]=true;
    cset[c.first][x]=true;
    sset[s.first][x]=true;
    restrict_rs[r.first][x]=true;
    restrict_cs[c.first][x]=true;
    restrict_sr[s.first][x]=true;
    restrict_sc[s.first][x]=true;
}

game(std::vector<std::string> const & desc)
{
    unsigned int k = 0;
    initialize();
    for(unsigned int i = 0 ; i < 9 ; i++) {
        for(unsigned int j = 0 ; j < 9 ; j++, k++) {
            if (desc[i][j] == '.') continue;
            unsigned int v = (unsigned char) (desc[i][j] - '1');
            set_pos(k, v);
        }
    }
}

void print()
{
    unsigned int k = 0;
    unsigned int v[81];
    for(unsigned int i = 0 ; i < npos ; i++) {
        v[pos[i]] = val[i];
    }
    for(unsigned int i = 0 ; i < 9 ; i++) {
        if (i == 3 || i == 6)
            cout << "\n";
        for(unsigned int j = 0 ; j < 9 ; j++, k++) {
            if (j == 3 || j == 6)
                cout << ' ';
            if (is_set[k]) {
                cout << (1+v[k]);
            } else {
                cout << '.';
            }
        }
        cout << "\n";
    }
#if 0
    cout << "\n";
    cout << npos << "\n";
    for(unsigned int i = 0 ; i < npos ; i++) {
        cout << pos[i] / 9 << " " << pos[i] % 9 << " " << val[i] << "\n";
    }
    cout << "\n";
#endif
}


int deduction();
void run(unsigned int, unsigned int);
void do_guess(unsigned int, unsigned int);
int saturate_groups();
int double_edges_search_general();

#if 0
int surely_in_one_of(unsigned int w, unsigned int i0, unsigned int i1)
{
    int moved=0;

    std::pair<unsigned int, unsigned int> r0,c0,s0;
    std::pair<unsigned int, unsigned int> r1,c1,s1;
    rcs_fromcell(r0,c0,s0,i0);
    rcs_fromcell(r1,c1,s1,i1);

    if (r0.first == r1.first) {
        for(unsigned int j = 0 ; j < 9 ; j++) {
            if (j == r0.second || j == r1.second)
                continue;
            moved += (ppl_row[r0.first][w] >> j) & 1;
            ppl_row[r0.first][w] &= ~(1 << j);
        }
    } else if (c0.first == c1.first) {
        for(unsigned int j = 0 ; j < 9 ; j++) {
            if (j == c0.second || j == c1.second)
                continue;
            moved += (ppl_col[c0.first][w] >> j) & 1;
            ppl_col[c0.first][w] &= ~(1 << j);
        }
    } else if (s0.first == s1.first) {
        for(unsigned int j = 0 ; j < 9 ; j++) {
            if (j == s0.second || j == s1.second)
                continue;
            moved += (ppl_sqr[s0.first][w] >> j) & 1;
            ppl_sqr[s0.first][w] &= ~(1 << j);
        }
    }

    return moved;
}
#endif

};

#if 0
int game::double_edges_search()
{
    uint16_t edges[81][81];
    int moved = 0;

    memset(edges,0,sizeof(edges));
    for(unsigned int i = 0 ; i < 9 ; i++) {
        uint16_t v = 1 << i;
        for(unsigned int z = 0 ; z < 9 ; z++) {
            unsigned int w;
            unsigned int t;
            t = ppl_row[z][i];
            w = __builtin_popcount(t);
            if (w == 2) {
                unsigned int a = __builtin_ffs(t) - 1;
                unsigned int b = __builtin_ffs(t ^ (1 << a)) - 1;
                edges[cell_fromrow(z,a)][cell_fromrow(z,b)]^=v;
            }
            t = ppl_col[z][i];
            w = __builtin_popcount(t);
            if (w == 2) {
                unsigned int a = __builtin_ffs(t) - 1;
                unsigned int b = __builtin_ffs(t ^ (1 << a)) - 1;
                edges[cell_fromcol(z,a)][cell_fromcol(z,b)]^=v;
            }
            t = ppl_sqr[z][i];
            w = __builtin_popcount(t);
            if (w == 2) {
                unsigned int a = __builtin_ffs(t) - 1;
                unsigned int b = __builtin_ffs(t ^ (1 << a)) - 1;
                edges[cell_fromsqr(z,a)][cell_fromsqr(z,b)]^=v;
            }
        }
    }
    for(unsigned int i = 0 ; i < 81 ; i++) {
        for(unsigned int j = 0 ; j < 81 ; j++) {
            unsigned int t = edges[i][j];
            unsigned int w = __builtin_popcount(t);
            if (w == 0)
                continue;
            unsigned int a = __builtin_ffs(t)-1;
            t ^= 1 << a;
            if (w == 1) {
                printf("x-wing for %u in rows %u and %u,cols %u and %u\n",
                        1+a,1+i/9,1+j/9,1+i%9,1+j%9);
                continue;
            }
            if (w > 2) {
                std::ostringstream o;
                o << (1+a) << " ";
                for(;t;) {
                    a = __builtin_ffs(t)-1;
                    t ^= 1 << a;
                    o << (1+a) << " ";
                }
                o << "cannot fit in edges " << (1+i) << " " << (1+j);
                throw std::logic_error(o.str());
            }
            unsigned int b = __builtin_ffs(t)-1;
            printf("found double edge %u and %u in %u and %u\n",1+a,1+b,1+i,1+j);

            moved += surely_in_one_of(a, i, j);
            moved += surely_in_one_of(b, i, j);
        }
    }
    printf("Killed %u possibilities from double-edges search\n", moved);
    return moved;
}
#endif

#if 0
// useless rubbish. And possibly buggy.


/* Assuming z gives the columns crossing row i, return a mask matching
 * the _squares_ crossed by row i
 */
static inline uint16_t xsqrs_fromrow(uint16_t z, uint16_t i)
{
    z = z | (z >> 1) | (z << 1);
    z = z & 0x92;
    // only bits 1 4 7 may be set
    z = (z >> 1) | (z >> 3) | (z >> 5);
    // set bits 0 1 2 (interesting) as well as 3 6 4, which are ignored.
    z = z & 0x07;
    z = z << (3 * (i / 3));
    return z;
}
static inline uint16_t xsqrs_fromcol(uint16_t z, uint16_t i)
{
    z = z | (z >> 1) | (z << 1);
    z = z & 0x92;
    // only bits 1 4 7 may be set
    z = z >> 1;
    z = z << (i % 3);
    return z;
}
static inline uint16_t xrows_fromcol(uint16_t z, uint16_t i) { return z; }
static inline uint16_t xcols_fromrow(uint16_t z, uint16_t i) { return z; }
static inline uint16_t xrows_fromsqr(uint16_t z, uint16_t i)
{
    return xsqrs_fromrow(z,i);
}
static inline uint16_t xcols_fromsqr(uint16_t z, uint16_t i)
{
    return xcols_fromrow(z,i);
}
// z gives a bit field marking columns. Give what would go in ppl_row[i] for
// indicating this.
static inline uint16_t cols_inrow(uint16_t z, uint16_t i) { return z; }
static inline uint16_t rows_incol(uint16_t z, uint16_t i) { return z; }
static inline uint16_t sqrs_inrow(uint16_t z, uint16_t i) {
    unsigned int s = 3 * (i / 3);
    z = z & (0x7 << s);
    z = z >> s;
    // then expand this, every bit will spread 3 positions.
    z = z | z << 2 | z << 4;
    z = z & 0x49;
    z = z | z << 1 | z << 2;
    return z;
}
static inline uint16_t sqrs_incol(uint16_t z, uint16_t i) {
    z = z & (0x49 << (i % 3));
    // some of either 036 147 or 258 are set.
    z = z | z >> 1 | z << 1;
    z = z >> 1;
    z = z & 0x49;
    z = (z << 3) - z;
    return z;
}
static inline uint16_t rows_insqr(uint16_t z, uint16_t i) {
    return sqrs_inrow(z,i);
}
static inline uint16_t cols_insqr(uint16_t z, uint16_t i) {
    return sqrs_incol(z,i);
}
#endif
static inline uint16_t identity(uint16_t z, unsigned int i __attribute__((unused))) {
    return z;
}

static inline unsigned int cell_rc(unsigned int i, unsigned int j)
{
    return i*9+j;
}

static inline unsigned int cell_cr(unsigned int j, unsigned int i)
{
    return i*9+j;
}

int game::saturate_groups()
{
    // check whether within the same row, some N digits can only fit in N
    // places. If this is the case, then other digits can be erased from
    // these places.
    //
    // it merely requires sorting ppl_row appropriately...
    int moved = 0;
#define SATURATE(ppl, s1, s2, cell) do {				\
    for(unsigned int i = 0 ; i < 9 ; i++) {				\
        typedef vector<pair<uint16_t,unsigned int> > tups_t;		\
        tups_t tups;							\
        for(unsigned int k = 0 ; k < 9 ; k++) {				\
            uint16_t z = ppl[i][k];					\
            /* If we have a power of two here,				\
             * then there's not much to do.				\
             */								\
            if ((z & (z - 1)) == 0)					\
                continue;						\
            tups.push_back(make_pair(z,k));	                        \
        }								\
        sort(tups.begin(),tups.end());					\
        for(tups_t::iterator u = tups.begin() ; u != tups.end();) {	\
            uint16_t z = u->first;					\
            unsigned int w = __builtin_popcount(z);			\
            unsigned int n = 0;						\
            ostringstream oo;						\
            uint16_t nz = 0;						\
            for( ; u != tups.end() && u->first == z ; u++, n++) {	\
                nz ^= 1 << u->second;					\
                oo << " " << (u->second+1);				\
            }								\
            if (n < w)							\
                continue;						\
            if (n > w) {						\
                std::ostringstream o;					\
                o << oo.str() << " cannot fit in " s2 "s"               \
                    << " " << allbits(z)                                \
                    << " for " s1 << (1+i) << "\n";			\
                throw std::logic_error(o.str());			\
            }								\
            /* WIN ! */							\
            uint16_t z0 = z;						\
            int moved_here = 0;						\
            for(unsigned int s = 0 ; z ; s++) {				\
                unsigned int j = __builtin_ffs(z)-1;			\
                z ^= 1 << j;						\
                moved_here |= cell_whofits[cell(i,j)] & ~nz;		\
                cell_whofits[cell(i,j)] &= nz;				\
            }								\
            if (moved_here) {						\
                cout << "{" << oo.str() << " } must be in " s2 "s"	\
                    << " " << allbits(z0)				\
                    << " for " s1 << " " << (1+i) << "\n";			\
                moved = 1;						\
            }								\
        }								\
    }                                                                   \
    /* we're not even sorting here...  it's still not			\
     * enough, since we have no guarantee that the covering		\
     * set is one of the values of the whofits[] array...		\
     *									\
     * for an order 2 saturation, clearly the whofits[]			\
     * values must be equal.						\
     *									\
     * for an order-3, it's already harder, since we may		\
     * have things such as [12][23][31] or [123][23][31],		\
     * for instance.							\
     *									\
     * for order-4, it's even worse, obviously.				\
     */									\
    for(unsigned int i = 0 ; i < 9 ; i++) {				\
        unsigned int wmax = 0;						\
        uint16_t excl = 0;						\
        for(unsigned int j = 0 ; j < 9 ; j++) {				\
            wmax += !rset[i][j];					\
            excl |= rset[i][j] << j;					\
        }								\
        /* the covering set may not exceed 2^wmax. */			\
        for(uint16_t z = 0 ; z < 512 ; z++) {				\
            /* If we have a power of two here,				\
             * then there's not much to do.				\
             */								\
            if (z & excl)						\
                continue;						\
            unsigned int w = __builtin_popcount(z);			\
            if ((z & (z - 1)) == 0)					\
                continue;						\
            ostringstream oo;						\
            unsigned int n = 0;						\
            uint16_t nz = 0;						\
            for(unsigned int jj = 0 ; jj < 9 ; jj++) {			\
                if (cell_whofits[cell(i,jj)] & ~z)			\
                    continue;						\
                nz ^= 1 << jj;						\
                oo << " " << (jj+1);					\
                n++;							\
            }								\
            /* we've found w values which have to fit within n cells. */\
            if (n < w)							\
                continue;						\
            if (n > w) {						\
                std::ostringstream o;					\
                o << allbits(z) << " cannot fit in " << s2 << "s"	\
                    << " " << oo.str()					\
                    << " for " << s1 << " " << (1+i) << "\n";		\
                throw std::logic_error(o.str());			\
            }								\
            /* WIN ! */							\
            uint16_t zz = z;						\
            int moved_here = 0;						\
            for(unsigned int s = 0 ; zz ; s++) {			\
                unsigned int k = __builtin_ffs(zz)-1;			\
                zz ^= 1 << k;						\
                moved_here |= ppl[i][k] & ~nz;	        		\
                ppl[i][k] &= nz;					\
            }								\
            if (moved_here) {						\
                cout << "{ " << allbits(z) << " } must be in "           \
                    << s2 << "s"                                        \
                    << oo.str()					\
                    << " for " << s1 << " " << (1+i) << "\n";		\
                moved = 1;						\
            }								\
        }								\
    }									\
    } while (0)
    SATURATE(ppl_row, "row", "col", cell_rc);
    SATURATE(ppl_col, "col", "row", cell_cr);
#if 1
#endif
    return moved;
}

int game::double_edges_search_general()
{
    // For each number x, look whether there are N rows with exactly N
    // *common* possibilities for placing x.
    //
    // It's terrilbly easy to check by sorting the ppl_row array !

    int moved = 0;

    // print();

    for(unsigned int k = 0 ; k < 9 ; k++) {
#define XWING(s1, s2, ppl1, ppl2, tr1, tr2)				\
        do {								\
            /* rows crossing cols */					\
            typedef vector<pair<uint16_t,unsigned int> > tups_t;	\
            tups_t tups;						\
            for(unsigned int i = 0 ; i < 9 ; i++) {			\
                uint16_t z = tr1(ppl1[i][k],i);				\
                /* If we have a power of two here,			\
                 * then there's not much to do.				\
                 */                                                     \
                if ((z & (z - 1)) == 0)					\
                    continue;						\
                tups.push_back(make_pair(z,i));	        		\
            }								\
            sort(tups.begin(),tups.end());				\
            for(tups_t::iterator u = tups.begin() ; u != tups.end();) { \
                uint16_t z = u->first;					\
                unsigned int w = __builtin_popcount(z);			\
                unsigned int n = 0;					\
                ostringstream oo;					\
                uint16_t nz = 0;					\
                for( ; u != tups.end() && u->first == z ; u++, n++) {	\
                    nz ^= 1 << u->second;				\
                    oo << " " << (u->second+1);				\
                }							\
                if (n < w)						\
                    continue;						\
                if (n > w) {						\
                    std::ostringstream o;				\
                    o << (1+k) << " cannot go in " s2 << allbits(z)	\
                        << " for " s1 << oo.str() << "\n";		\
                    throw std::logic_error(o.str());			\
                }							\
                /* WIN ! */						\
                uint16_t z0 = z;					\
                int moved_here = 0;					\
                for(unsigned int s = 0 ; z ; s++) {			\
                    unsigned int a = __builtin_ffs(z)-1;		\
                    z ^= 1 << a;					\
                    uint16_t nz2 = tr2(nz, a);				\
                    moved_here |= ppl2[a][k] & ~nz2;                    \
                    ppl2[a][k] &= nz2;					\
                }							\
                if (moved_here) {					\
                    cout << "Order-" << n << " x-wing for " << (1+k)	\
                        << " in " s1 << oo.str()			\
                        << " ; killing rest in " s2 " " << allbits(z0)	\
                        << "\n";					\
                    moved = 1;						\
                }							\
            }								\
        } while (0)
        XWING("rows", "cols", ppl_row, ppl_col, identity, identity);
        XWING("cols", "rows", ppl_col, ppl_row, identity, identity);
        // XWING("rows", "cols", ppl_row, ppl_col, xcols_fromrow, rows_incol);
        // XWING("cols", "rows", ppl_col, ppl_row, xrows_fromcol, cols_inrow);
        // This cannot have any interest. If two rows hit the same two
        // squares, then the corresponding digit has to go in the
        // complementary row in the complementary square, anyway --
        // because it's known not to fit anywhere else within that
        // square.
        // XWING("rows", "sqrs", ppl_row, ppl_sqr, xsqrs_fromrow, rows_insqr);
        // XWING("cols", "sqrs", ppl_col, ppl_sqr, xsqrs_fromcol, cols_insqr);
        // conversely, these two cannot have any interest either, from
        // the same argument.
        // XWING("sqrs", "rows", ppl_sqr, ppl_row, xrows_fromsqr, sqrs_inrow);
        // XWING("sqrs", "cols", ppl_sqr, ppl_col, xcols_fromsqr, sqrs_incol);
    }
    return moved;
}

int game::deduction()
{
    int moved=0;

    // print();
    printf("There are %d cells set\n", npos);
    /* Loop through cells */
    for(unsigned int i = 0 ; i < 81 ; i++) {
        if (is_set[i]) continue;

        uint16_t v = cell_whofits[i];

        std::pair<unsigned int, unsigned int> r,c,s;
        rcs_fromcell(r,c,s,i);

        /* we take the ``whofits'' information so as to reduce the number of
         * possibilities in the ppl (possible placements) arrays */
        /* We also to the converse, since both might impact eachother */

        for(unsigned int k = 0 ; k < 9 ; k++) {

            unsigned int a = v >> k;
            unsigned int xr = ppl_row[r.first][k] >> r.second;
            unsigned int xc = ppl_col[c.first][k] >> c.second;
            unsigned int xs = ppl_sqr[s.first][k] >> s.second;

            if (!(a & xr & xc & xs  & 1UL)) {
                moved += (a | xr | xc | xs) & 1;

                cell_whofits[i] &= ~(1 << k);
                ppl_row[r.first][k] &= ~(1 << r.second);
                ppl_col[c.first][k] &= ~(1 << c.second);
                ppl_sqr[s.first][k] &= ~(1 << s.second);
            }
        }

        if (cell_whofits[i] == 0) {
            std::ostringstream o;
            o << "position " << (1+i) << " cannot be filled";
            throw std::logic_error(o.str());
        }

        if (is_bit(v)) {
            unsigned int k = __builtin_ffs(v)-1;
            printf("cell %u (row %u col %u sqr %u -- idx %u) may only contain a %u\n", 1+i, 1+r.first, 1+c.first, 1+s.first, s.second, k+1);
            set_pos(i,k);
            moved++;
            continue;
        }

    }

    // printf("There are %d cells set\n", npos);
    /* Loop through rows, to see where integers will be put */
    for(unsigned int i = 0 ; i < 9 ; i++) {
        for(unsigned int k = 0 ; k < 9 ; k++) {
            if (rset[i][k]) continue;
            uint16_t v = ppl_row[i][k];

            if (!v) {
                std::ostringstream o;
                o << (1+k) << " can go nowhere in row " << (1+i);
                throw std::logic_error(o.str());
            }

            if (is_bit(v)) {
                unsigned int j = __builtin_ffs(v)-1;
                printf("In row %u, the %u must go to position %u\n",
                        1+i, k+1, 1+j);
                set_pos(cell_fromrow(i,j), k);
                moved++;
            } else if (!restrict_rs[i][k]) {
                uint16_t m = 0x7;
                unsigned int i1 = i % 3;
                uint16_t m1 = m << (i1*3);
                for(unsigned int l = 0 ; l < 3 ; l++) {
                    if (v == (v & m)) {
                        printf("In square %u, the %u comes from row %u\n",
                                1+i-i1 + l, k+1, 1+i);
                        /* So forbid it elsewhere */
                        ppl_sqr[i-i1+l][k] &= m1;
                        moved++;
                        restrict_rs[i][k]=1;
                    }
                    m <<= 3;
                }
            }
        }
    }

    // printf("There are %d cells set\n", npos);
    /* Loop through cols, to see where integers will be put */
    for(unsigned int j = 0 ; j < 9 ; j++) {
        for(unsigned int k = 0 ; k < 9 ; k++) {
            if (cset[j][k]) continue;
            uint16_t v = ppl_col[j][k];
            if (!v) {
                std::ostringstream o;
                o << (1+k) << " can go nowhere in col " << 1+j;
                throw std::logic_error(o.str());
            }


            if (is_bit(v)) {
                unsigned int i = __builtin_ffs(v)-1;
                printf("In col %u, the %u must go to position %u\n",
                        1+j, k+1, 1+i);
                set_pos(cell_fromcol(j,i), k);
            } else if (!restrict_cs[j][k]) {
                /* Intersecting squares are j/3, j/3+3, j/3+6 */
                uint16_t m = 0x7;
                unsigned int jq = j / 3;
                unsigned int jr = j % 3;
                uint16_t m1 = 0x49 << jr; // 0x49 is 0 0100 1001
                for(unsigned int l = 0 ; l < 9 ; l+= 3) {
                    if (v == (v & m)) {
                        printf("In square %u, the %u comes from col %u\n",
                                1+jq + l, k+1, 1+j);
                        /* So forbid it elsewhere */
                        ppl_sqr[jq+l][k] &= m1;
                        moved++;
                        restrict_cs[j][k]=1;
                    }
                    m <<= 3;
                }
            }
        }
    }

    // printf("There are %d cells set\n", npos);
    /* Loop through squares, to see where integers will be put */
    for(unsigned int s = 0 ; s < 9 ; s++) {
        for(unsigned int k = 0 ; k < 9 ; k++) {
            if (sset[s][k]) continue;
            uint16_t v = ppl_sqr[s][k];
            if (!v) {
                std::ostringstream o;
                o << (1+k) << " can go nowhere in sqr " << (1+s);
                throw std::logic_error(o.str());
            }

            if (is_bit(v)) {
                unsigned int i = __builtin_ffs(v)-1;
                printf("In sqr %u, the %u must go to position %u\n",
                        1+s, k+1, 1+i);
                set_pos(cell_fromsqr(s,i), k);
            } else {
                unsigned int s1 = s%3;
                unsigned int s0 = s - s1;
                unsigned int s2 = 3*s1;
                uint16_t m,m1;

                if (!restrict_sr[s][k]) {
                    /* intersecting rows are s0, s0+1, s0+2 */
                    /* Our mask for these rows is 0x7 shifted 3s1 times */
                    m = 0x7;
                    m1 = m << s2;
                    for(unsigned int l = 0 ; l < 3 ; l++) {
                        if (v == (v & m)) {
                            printf("In row %u, the %u comes from square %u\n",
                                    1+s0 + l, k+1, 1+s);
                            /* So forbid it elsewhere */
                            ppl_row[s0+l][k] &= m1;
                            moved++;
                            restrict_sr[s][k]=1;
                        }
                        m <<= 3;
                    }
                }

                if (!restrict_sc[s][k]) {
                    /* intersecting cols are s2, s2+1, s2+2 */
                    /* our mask is 0x7 shifted s0 times */
                    m = 0x49;
                    m1 = 0x7 << s0;
                    for(unsigned int l = 0 ; l < 3 ; l++) {
                        if (v == (v & m)) {
                            printf("In col %u, the %u comes from square %u\n",
                                    1+s2 + l, k+1, 1+s);
                            /* So forbid it elsewhere */
                            ppl_col[s2+l][k] &= m1;
                            moved++;
                            restrict_sc[s][k]=1;
                        }
                        m <<= 1;
                    }
                }
            }
        }
    }

    if (moved == 0)
        moved = saturate_groups();

    // it turns out that the x-wing search is faster with
    // the present implementation, even though it appears
    // less ``visual'', and less ``human accessible''.
    if (moved == 0)
        moved = double_edges_search_general();


    return moved;
}
void game::run(unsigned int nguesses, unsigned int nbacked)
{
    for(;npos < 81 && deduction();) ;
    if (npos == 81) {
        solcount++;
        printf("Solution found (total %u -- %u guesses, %u backed)\n", solcount,
                nguesses, nbacked);
        print();
        return;
    }
    printf("No more moves with logic\n");
    if (no_guess) {
        print();
        printf("\n");
        unsigned int k = 0;
        for(unsigned int i = 0 ; i < 9 ; i++) {
                if (i == 3 || i == 6)
                    cout << "\n";
            for(unsigned int j = 0 ; j < 9 ; j++) {
                if (j == 3 || j == 6)
                    cout << "\t";
                else if (j)
                    cout << " ";
                cout << allbits(cell_whofits[k],"");
                k++;
            }
            cout << "\n";
        }
        exit(1);
    }
    do_guess(nguesses, nbacked);
}
void game::do_guess(unsigned int nguesses, unsigned int nbacked)
{
    /* Find the cell with the fewer number of possibilities, this will
     * give us some ordering. Actually, we would prefer to draw some sort
     * of graph, and focus on the verices of largest degree */
    unsigned int lightest=UINT_MAX;
    unsigned int which=UINT_MAX;

    for(unsigned int i = 0 ; i < 81 ; i++) {
        if (is_set[i]) continue;
        unsigned int w = __builtin_popcount(cell_whofits[i]);
        if (w < lightest) {
            lightest = w;
            which = i;
        }
    }

    printf("Starting to guess ; situation deduced so far:\n");
    print();

    uint16_t v = cell_whofits[which];
    unsigned int nt=0;
    for(unsigned int j = 0 ; j < 9 ; j++, v>>=1) {
        if (!(v & 1UL))
            continue;

        nt++;

        printf("<<< [%u] guess %u in cell %u (# %u/%u)\n",
                nguesses, 1+j, 1+which, nt, lightest);

        try {
            if (nt == lightest) {
                // then we might as well save a copy !
                // normally set_pos is unnecessary.
                assert(is_bit(cell_whofits[which]));
                // set_pos(which, j);
                run(nguesses + 1, nbacked);
            } else {
                game g1(*this);
                g1.set_pos(which, j);
                g1.run(nguesses + 1, nbacked);
            }
        } catch (std::logic_error x) {
            printf(">>> [%u] guess # %u/%u was wrong [%s]\n",
                    nguesses, nt, lightest, x.what());
        }
        printf("--- [%u] now assuming that cell %u cannot be a %u\n",
                nguesses, 1+which, 1+j);
        forbid_pos(which, j);

        try {
            if (deduction()) {
                printf("--- [%u] our main position has advanced, going back to logic loop\n", nguesses);
                run(nguesses+1,nbacked+1);
                return;
            }
        } catch (std::logic_error x) {
            printf(">>> [%u] un-guess # %u/%u was wrong [%s]\n",
                    nguesses, nt, lightest, x.what());
        }
    }
}

int main(int argc, char * argv[])
{
    std::vector<std::string> desc;

    if (argc > 2 && strcmp(argv[1], "--no-guesses") == 0) {
        argc--,argv++;
        no_guess = 1;
    }

    if (argc != 2) {
        cerr << "Usage: ./solver <input file>\n";
        exit(1);
    }

    ifstream in(argv[1]);
    input_pb(desc, in);
    assert(desc.size() == 9);
    // print_desc(desc);


    game g(desc);

    g.run(0,0);

    printf("%u solutions found\n", solcount);

    return 0;
}
